

from typing import Callable, List
from gaia_sdk.api.VariableRegistry import VariableRegistry
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType


class CreateSkillProvisionImpulse():
    """
    The specification to create a SkillProvision instance
    """
    tenantId: str
    qualifier: str
    appendent: str
    labelList: List[str]
    version: str
    skillRef: str
    cpu: int
    memory: int
    replicas: int
    enabled: bool
    bootstrapTimeout: int
    environment: dict

    def __init__(self, tenantId: str, qualifier: str, appendent: str, labelList: List[str], version: str, skillRef: str, cpu: int, memory: int, replicas: int, enabled: bool, bootstrapTimeout: int, environment: dict):
        self.tenantId = tenantId
        self.qualifier = qualifier
        self.appendent = appendent
        self.labelList = labelList
        self.version = version
        self.skillRef = skillRef
        self.cpu = cpu
        self.memory = memory
        self.replicas = replicas
        self.enabled = enabled
        self.bootstrapTimeout = bootstrapTimeout
        self.environment = environment

    def __eq__(self, other):
        if type(other) is type(self):
            return self.tenantId == other.tenantId and self.qualifier == other.qualifier and self.appendent == other.appendent and self.labelList == other.labelList and self.version == other.version and self.skillRef == other.skillRef and self.cpu == other.cpu and self.memory == other.memory and self.replicas == other.replicas and self.enabled == other.enabled and self.bootstrapTimeout == other.bootstrapTimeout and self.environment == other.environment
        return False

    def __repr__(self):
        return {'tenantId': self.tenantId, 'qualifier': self.qualifier, 'appendent': self.appendent, 'labelList': self.labelList, 'version': self.version, 'skillRef': self.skillRef, 'cpu': self.cpu, 'memory': self.memory, 'replicas': self.replicas, 'enabled': self.enabled, 'bootstrapTimeout': self.bootstrapTimeout, 'environment': self.environment}




from gaia_sdk.graphql.response.type.CreatedEdgeImpulse import CreatedEdgeImpulse
from gaia_sdk.graphql.response.type.CreatedCodeImpulse import CreatedCodeImpulse
from gaia_sdk.graphql.response.type.CreatedPromptImpulse import CreatedPromptImpulse
from gaia_sdk.graphql.response.type.CreatedIntentImpulse import CreatedIntentImpulse
from gaia_sdk.graphql.response.type.CreatedSkillImpulse import CreatedSkillImpulse
from gaia_sdk.graphql.response.type.CreatedFulfilmentImpulse import CreatedFulfilmentImpulse
from gaia_sdk.graphql.response.type.CreatedUserImpulse import CreatedUserImpulse
from gaia_sdk.graphql.response.type.CreatedSkillProvisionImpulse import CreatedSkillProvisionImpulse
from gaia_sdk.graphql.response.type.CreatedTenantImpulse import CreatedTenantImpulse
from gaia_sdk.graphql.response.type.CreatedStatementImpulse import CreatedStatementImpulse
from gaia_sdk.graphql.response.type.CreatedRoleImpulse import CreatedRoleImpulse
from gaia_sdk.graphql.response.type.CreatedBehaviourImpulse import CreatedBehaviourImpulse
from gaia_sdk.graphql.response.type.CreatedIdentityImpulse import CreatedIdentityImpulse
from gaia_sdk.graphql.response.type.CreatedApiKeyImpulse import CreatedApiKeyImpulse
from gaia_sdk.graphql.request.input.CreateTenantImpulse import CreateTenantImpulse
from gaia_sdk.graphql.request.input.CreateSkillImpulse import CreateSkillImpulse
from gaia_sdk.graphql.request.input.CreateEdgeImpulse import CreateEdgeImpulse
from gaia_sdk.graphql.request.input.CreateIdentityImpulse import CreateIdentityImpulse
from gaia_sdk.graphql.request.input.CreateCodeImpulse import CreateCodeImpulse
from gaia_sdk.graphql.request.input.CreateFulfilmentImpulse import CreateFulfilmentImpulse
from gaia_sdk.graphql.request.input.CreateStatementImpulse import CreateStatementImpulse
from gaia_sdk.graphql.request.input.CreateSkillProvisionImpulse import CreateSkillProvisionImpulse
from gaia_sdk.graphql.request.input.CreateRoleImpulse import CreateRoleImpulse
from gaia_sdk.graphql.request.input.CreateIntentImpulse import CreateIntentImpulse
from gaia_sdk.graphql.request.input.CreatePromptImpulse import CreatePromptImpulse
from gaia_sdk.graphql.request.input.CreateBehaviourImpulse import CreateBehaviourImpulse
from gaia_sdk.graphql.request.input.CreateUserImpulse import CreateUserImpulse
from gaia_sdk.graphql.request.input.CreateApiKeyImpulse import CreateApiKeyImpulse

from typing import List
Uuid = str
String = str
ISO8601 = str
Struct = dict
Float = float
Int = int
Boolean = bool
from gaia_sdk.graphql.request.enumeration.RuntimeState import RuntimeState
from gaia_sdk.graphql.request.enumeration.SkillState import SkillState
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType

class CreateKnowledge:
    dictionary: dict

    def __init__(self, dictionary: dict):
        self.dictionary = dictionary

    def __eq__(self, other):
        if type(other) is type(self):
            return self.dictionary == other.dictionary
        return False

    def __repr__(self):
        return {'dictionary': self.dictionary}

    """
    creates a list of identities with the given specifications
    """
    @property
    def identities(self) -> List[CreatedIdentityImpulse]:
        return list(map(lambda x: CreatedIdentityImpulse(x), self.dictionary.get("identities")))
    """
    creates a list of tenants with the given specifications
    """
    @property
    def tenants(self) -> List[CreatedTenantImpulse]:
        return list(map(lambda x: CreatedTenantImpulse(x), self.dictionary.get("tenants")))
    """
    creates a list of users with the given specifications
    """
    @property
    def users(self) -> List[CreatedUserImpulse]:
        return list(map(lambda x: CreatedUserImpulse(x), self.dictionary.get("users")))
    """
    creates a list of api keys with the given specifications
    """
    @property
    def api_keys(self) -> List[CreatedApiKeyImpulse]:
        return list(map(lambda x: CreatedApiKeyImpulse(x), self.dictionary.get("apiKeys")))
    """
    creates a list of roles with the given specifications
    """
    @property
    def roles(self) -> List[CreatedRoleImpulse]:
        return list(map(lambda x: CreatedRoleImpulse(x), self.dictionary.get("roles")))
    """
    creates a list of intents with the given specifications
    """
    @property
    def intents(self) -> List[CreatedIntentImpulse]:
        return list(map(lambda x: CreatedIntentImpulse(x), self.dictionary.get("intents")))
    """
    creates a list of prompts with the given specifications
    """
    @property
    def prompts(self) -> List[CreatedPromptImpulse]:
        return list(map(lambda x: CreatedPromptImpulse(x), self.dictionary.get("prompts")))
    """
    creates a list of statements with the given specifications
    """
    @property
    def statements(self) -> List[CreatedStatementImpulse]:
        return list(map(lambda x: CreatedStatementImpulse(x), self.dictionary.get("statements")))
    """
    creates a list of fulfilments with the given specifications
    """
    @property
    def fulfilments(self) -> List[CreatedFulfilmentImpulse]:
        return list(map(lambda x: CreatedFulfilmentImpulse(x), self.dictionary.get("fulfilments")))
    """
    creates a list of behaviours with the given specifications
    """
    @property
    def behaviours(self) -> List[CreatedBehaviourImpulse]:
        return list(map(lambda x: CreatedBehaviourImpulse(x), self.dictionary.get("behaviours")))
    """
    creates a list of codes with the given specifications
    """
    @property
    def codes(self) -> List[CreatedCodeImpulse]:
        return list(map(lambda x: CreatedCodeImpulse(x), self.dictionary.get("codes")))
    """
    creates a list of edges with the given specifications
    """
    @property
    def edges(self) -> List[CreatedEdgeImpulse]:
        return list(map(lambda x: CreatedEdgeImpulse(x), self.dictionary.get("edges")))
    """
    creates a list of skills with the given specifications
    """
    @property
    def skills(self) -> List[CreatedSkillImpulse]:
        return list(map(lambda x: CreatedSkillImpulse(x), self.dictionary.get("skills")))
    """
    creates a list of skill provisions with the given specifications
    """
    @property
    def skill_provisions(self) -> List[CreatedSkillProvisionImpulse]:
        return list(map(lambda x: CreatedSkillProvisionImpulse(x), self.dictionary.get("skillProvisions")))



from abc import ABCMeta, abstractmethod

from requests import Response

from gaia_sdk.api.client_options import ClientOptions
from gaia_sdk.http.request.Payload import Payload


class ITransporter:
    """
    Abstract transporter implementation. Transporter classes are used to establish a connection to G.A.I.A.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def transport(self, options: ClientOptions, payload: Payload, url_post_fix: str = "") -> Response: raise NotImplementedError





from typing import Callable, List
from gaia_sdk.api.VariableRegistry import VariableRegistry
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType


class User(list):
    """
    Represents User information
    """

    """
    Id of the user
    """
    def user_id(self):
        self.append(lambda x: "userId")

    """
    The username of the user
    """
    def username(self):
        self.append(lambda x: "username")

    """
    The email of the user
    """
    def email(self):
        self.append(lambda x: "email")

    """
    The first name of the user
    """
    def first_name(self):
        self.append(lambda x: "firstName")

    """
    The last name of the user
    """
    def last_name(self):
        self.append(lambda x: "lastName")

    """
    Is the User using 2FA
    """
    def using2_f_a(self):
        self.append(lambda x: "using2FA")

    """
    The tenants of the user
    """
    def tenants(self):
        self.append(lambda x: "tenants")

    """
    The roles of the user
    """
    def roles(self):
        self.append(lambda x: "roles")

    """
    The groups of the user
    """
    def groups(self):
        self.append(lambda x: "groups")

    """
    The permissions of the user
    """
    def permissions(self):
        self.append(lambda x: "permissions")

    def render(self, registry: VariableRegistry):
        return " ".join(map(lambda e: e(registry), self))



import unittest

from rx import operators as ops, pipe
from gaia_sdk.api.GaiaCredentials import UsernamePasswordCredentials
from gaia_sdk.tests.mock import mock_gaia_ref
from gaia_sdk.tests.mock import MockResponse
from gaia_sdk.gaia import Gaia


class TestIdentityRef(unittest.TestCase):

    def test_export_identity_mock(self):
        def mock(request):
            self.assertEqual(request.url_post_fix, "/identity/source")
            return MockResponse(bytes("identity content", encoding="utf-8"))

        self.gaiaRef = mock_gaia_ref(mock)
        result = pipe(ops.first())(
            self.gaiaRef.identity().export("00000000-0000-0000-0000-000000000000")).run()
        self.assertEqual(result, bytes("identity content", "utf-8"))

    def test_export_identity_no_id(self):
        def mock(request):
            self.assertEqual(request.url_post_fix, "/identity/source")
            return MockResponse(bytes("identity content", encoding="utf-8"))

        self.gaiaRef = mock_gaia_ref(mock)
        self.assertRaises(TypeError, lambda: pipe(ops.first())(self.gaiaRef.identity().export()).run())

    @unittest.skip("E2E test, for local use or future E2E use")
    def test_export_identity_e2e(self):
        self.gaiaRef = Gaia.login("http://localhost:8080", UsernamePasswordCredentials("admin", "admin"))

        result = pipe(ops.first())(
            self.gaiaRef.identity().export("d32829c8-5900-4346-9577-25e8146d1e78")).run()
        self.assertEqual(len(result), 28311)

    def test_import_identity(self):
        self.gaiaRef = mock_gaia_ref(self.mock_write)
        data = bytes("identity content", encoding="utf-8")
        result = pipe(ops.first())(self.gaiaRef.identity().import_identity('tenantId', 'identityName', data)).run()
        self.assertEqual(result['uri'], 'gaia://tenantId/identities/identityName')
        self.assertEqual(result['partitionKey'], '0123456789')

    @unittest.skip("E2E test, for local use or future E2E use")
    def test_import_identity_e2e(self):
        self.gaiaRef = Gaia.login("http://localhost:8080", UsernamePasswordCredentials("admin", "admin"))
        data = bytes("identity content", encoding="utf-8")
        result = pipe(ops.first())(
            self.gaiaRef.identity().import_identity('2fa4ff18-5c30-497b-9ad2-0d8eb51cd4da',
                                                    'identityName', data, True)).run()
        self.assertEqual(len(result.uri), 72)

        result = pipe(ops.first())(
            self.gaiaRef.data("gaia://2fa4ff18-5c30-497b-9ad2-0d8eb51cd4da/").list()).run()
        self.assertEqual(result[0].dictionary, {'filePath': 'identities/identityName',
                                                'tenant': '2fa4ff18-5c30-497b-9ad2-0d8eb51cd4da'})
        pass

    def mock_write(self, request):
        if request.url_post_fix == "/data/sink/init":
            return MockResponse({"uploadId": "A123"})
        elif request.url_post_fix == "/data/sink/chunk":
            return MockResponse({"chunkId": "B123"})
        elif request.url_post_fix == "/data/sink/complete":
            return MockResponse("asdfqwer")
        elif request.url_post_fix == "/identity/import":
            return MockResponse({'partitionKey': '0123456789', 'uri': 'gaia://tenantId/identities/identityName'})
        else:
            raise RuntimeError('On purpose')



from abc import abstractmethod, ABC
from rx.core.typing import Observable
from typing import Callable, List

from gaia_sdk.graphql import RetrievalReq, ExperienceReq, KnowledgeReq, EdgeReq, \
    IdentityReq, IntentReq, PromptReq, StatementReq, FulfilmentReq, CodeReq, BehaviourReq, IntrospectionReq, SkillIntrospectionReq, \
    PerceptionReq, PreservationReq, CreatedIdentityImpulse, UpdatedIdentityImpulse, DeletedIdentityImpulse, \
    CreatedIntentImpulse, UpdatedIntentImpulse, DeletedIntentImpulse, RetrievalRes, \
    ExperienceRes, EdgeRes, StatementRes, PromptRes, IntentRes, IdentityRes, KnowledgeRes, FulfilmentRes, CodeRes, \
    BehaviourRes, IntrospectionRes, SkillIntrospectionRes, PreservationRes, PerceptionRes, PerceivedImpulse, \
    DeleteIdentityImpulse, UpdateIdentityImpulse, CreateIdentityImpulse, DeleteIntentImpulse, UpdateIntentImpulse, CreateIntentImpulse,  \
    PerceiveActionImpulse, PerceiveDataImpulse, \
    CreatePromptImpulse, UpdatePromptImpulse, DeletePromptImpulse, CreatedPromptImpulse, UpdatedPromptImpulse, \
    DeletedPromptImpulse, CreateStatementImpulse, UpdateStatementImpulse, DeleteStatementImpulse, \
    CreatedStatementImpulse, UpdatedStatementImpulse, DeletedStatementImpulse, CreateFulfilmentImpulse, \
    UpdateFulfilmentImpulse, DeleteFulfilmentImpulse, CreatedFulfilmentImpulse, UpdatedFulfilmentImpulse, \
    DeletedFulfilmentImpulse, CreateBehaviourImpulse, UpdateBehaviourImpulse, DeleteBehaviourImpulse, \
    CreatedBehaviourImpulse, UpdatedBehaviourImpulse, DeletedBehaviourImpulse, CreateCodeImpulse, UpdateCodeImpulse, \
    DeleteCodeImpulse, CreatedCodeImpulse, UpdatedCodeImpulse, DeletedCodeImpulse, CreateEdgeImpulse, \
    DeleteEdgeImpulse, CreatedEdgeImpulse, DeletedEdgeImpulse, BehaviourExecutionRes, BehaviourExecutionReq, \
    BehaviourExecutionDetailReq, BehaviourExecutionDetailRes, \
    SkillProvisionBuildJobReq, SkillProvisionBuildJobRes

Uuid = str

class ISensorFunction(ABC):

    @abstractmethod
    def retrieve(self, config: Callable[[RetrievalReq], None]) -> Observable[RetrievalRes]:
        pass

    @abstractmethod
    def retrieve_experience(self, config: Callable[[ExperienceReq], None]) -> Observable[ExperienceRes]:
        pass

    @abstractmethod
    def retrieve_knowledge(self, config: Callable[[KnowledgeReq], None]) -> Observable[KnowledgeRes]:
        pass

    @abstractmethod
    def retrieve_edges(self, source: Uuid, config: Callable[[EdgeReq], None], limit: int, offset: int) -> Observable[EdgeRes]:
        pass

    @abstractmethod
    def retrieve_edge(self, source: Uuid, edgeId: Uuid, config: Callable[[EdgeReq], None]) -> Observable[EdgeRes]:
        pass

    @abstractmethod
    def retrieve_identities(self, config: Callable[[IdentityReq], None], limit: int, offset: int) -> Observable[IdentityRes]:
        pass

    @abstractmethod
    def retrieve_identity(self, identity_id: Uuid, config: Callable[[IdentityReq], None]) -> Observable[IdentityRes]:
        pass

    @abstractmethod
    def retrieve_intents(self, identityId: Uuid, config: Callable[[IntentReq], None], limit: int, offset: int) -> Observable[IntentRes]:
        pass

    @abstractmethod
    def retrieve_intent(self, identity_id: Uuid, reference: Uuid, config: Callable[[IntentReq], None]) -> Observable[IntentRes]:
        pass

    @abstractmethod
    def retrieve_prompts(self, identity_id: Uuid, config: Callable[[PromptReq], None], limit: int, offset: int) -> Observable[PromptRes]:
        pass

    @abstractmethod
    def retrieve_prompt(self, identity_id: Uuid, reference: Uuid,  config: Callable[[PromptReq], None]) -> Observable[PromptRes]:
        pass

    @abstractmethod
    def retrieve_statements(self, identity_id: Uuid, config: Callable[[StatementReq], None], limit: int, offset: int) -> Observable[StatementRes]:
        pass

    @abstractmethod
    def retrieve_statement(self, identity_id: Uuid, reference: Uuid, config: Callable[[StatementReq], None]) -> Observable[StatementRes]:
        pass

    @abstractmethod
    def retrieve_fulfilments(self, identity_id: Uuid, config: Callable[[FulfilmentReq], None], limit: int, offset: int) -> Observable[FulfilmentRes]:
        pass

    @abstractmethod
    def retrieve_fulfilment(self, identity_id: Uuid, reference: Uuid, config: Callable[[FulfilmentReq], None]) -> Observable[FulfilmentRes]:
        pass

    @abstractmethod
    def retrieve_codes(self, identity_id: Uuid, config: Callable[[CodeReq], None], limit: int, offset: int) -> Observable[CodeRes]:
        pass

    @abstractmethod
    def retrieve_code(self, identity_id: Uuid, reference: Uuid, config: Callable[[CodeReq], None]) -> Observable[CodeRes]:
        pass

    @abstractmethod
    def retrieve_behaviours(self, identity_id: Uuid, config: Callable[[BehaviourReq], None], limit: int, offset: int) -> Observable[BehaviourRes]:
        pass

    @abstractmethod
    def retrieve_behaviour(self, identity_id: Uuid, reference: Uuid, config: Callable[[BehaviourReq], None]) -> Observable[BehaviourRes]:
        pass

    @abstractmethod
    def retrieve_behaviour_execution(self, identity_id: Uuid, process_instance_id: Uuid, config: Callable[[BehaviourExecutionDetailReq], None]) \
            -> Observable[BehaviourExecutionDetailRes]:
        pass

    @abstractmethod
    def retrieve_behaviour_executions(self, identity_id: Uuid, config: Callable[[BehaviourExecutionReq], None], limit: int = None, offset: int = None, startDate: str = None, endDate: str = None) \
            -> Observable[BehaviourExecutionRes]:
        pass

    @abstractmethod
    def retrieve_skill_provision_build_jobs(self, tenant_id: Uuid, config: Callable[[SkillProvisionBuildJobReq], None]) -> Observable[SkillProvisionBuildJobRes]:
        pass

    @abstractmethod
    def introspect(self, config: Callable[[IntrospectionReq], None]) -> Observable[IntrospectionRes]:
        pass

    @abstractmethod
    def introspect_skills(self, config: Callable[[SkillIntrospectionReq], None]) -> Observable[SkillIntrospectionRes]:
        pass

    @abstractmethod
    def preserve(self, config: Callable[[PreservationReq], None]) -> Observable[PreservationRes]:
        pass

    @abstractmethod
    def preserve_create_identities(self, impulses: List[CreateIdentityImpulse]) -> Observable[CreatedIdentityImpulse]:
        pass

    @abstractmethod
    def preserve_update_identities(self, impulses: List[UpdateIdentityImpulse]) -> Observable[UpdatedIdentityImpulse]:
        pass

    @abstractmethod
    def preserve_delete_identities(self, impulses: List[DeleteIdentityImpulse]) -> Observable[DeletedIdentityImpulse]:
        pass

    @abstractmethod
    def preserve_create_intents(self, impulses: List[CreateIntentImpulse]) -> Observable[CreatedIntentImpulse]:
        pass

    @abstractmethod
    def preserve_update_intents(self, impulses: List[UpdateIntentImpulse]) -> Observable[UpdatedIntentImpulse]:
        pass

    @abstractmethod
    def preserve_delete_intents(self, impulses: List[DeleteIntentImpulse]) -> Observable[DeletedIntentImpulse]:
        pass

    @abstractmethod
    def preserve_create_prompts(self, impulses: List[CreatePromptImpulse]) -> Observable[CreatedPromptImpulse]:
        pass

    @abstractmethod
    def preserve_update_prompts(self, impulses: List[UpdatePromptImpulse]) -> Observable[UpdatedPromptImpulse]:
        pass

    @abstractmethod
    def preserve_delete_prompts(self, impulses: List[DeletePromptImpulse]) -> Observable[DeletedPromptImpulse]:
        pass

    @abstractmethod
    def preserve_create_statements(self, impulses: List[CreateStatementImpulse]) -> Observable[CreatedStatementImpulse]:
        pass

    @abstractmethod
    def preserve_update_statements(self, impulses: List[UpdateStatementImpulse]) -> Observable[UpdatedStatementImpulse]:
        pass

    @abstractmethod
    def preserve_delete_statements(self, impulses: List[DeleteStatementImpulse]) -> Observable[DeletedStatementImpulse]:
        pass

    @abstractmethod
    def preserve_create_fulfilments(self, impulses: List[CreateFulfilmentImpulse]) -> Observable[CreatedFulfilmentImpulse]:
        pass

    @abstractmethod
    def preserve_update_fulfilments(self, impulses: List[UpdateFulfilmentImpulse]) -> Observable[UpdatedFulfilmentImpulse]:
        pass

    @abstractmethod
    def preserve_delete_fulfilments(self, impulses: List[DeleteFulfilmentImpulse]) -> Observable[DeletedFulfilmentImpulse]:
        pass

    @abstractmethod
    def preserve_create_behaviours(self, impulses: List[CreateBehaviourImpulse]) -> Observable[CreatedBehaviourImpulse]:
        pass

    @abstractmethod
    def preserve_update_behaviours(self, impulses: List[UpdateBehaviourImpulse]) -> Observable[UpdatedBehaviourImpulse]:
        pass

    @abstractmethod
    def preserve_delete_behaviours(self, impulses: List[DeleteBehaviourImpulse]) -> Observable[DeletedBehaviourImpulse]:
        pass

    @abstractmethod
    def preserve_create_codes(self, impulses: List[CreateCodeImpulse]) -> Observable[CreatedCodeImpulse]:
        pass

    @abstractmethod
    def preserve_update_codes(self, impulses: List[UpdateCodeImpulse]) -> Observable[UpdatedCodeImpulse]:
        pass

    @abstractmethod
    def preserve_delete_codes(self, impulses: List[DeleteCodeImpulse]) -> Observable[DeletedCodeImpulse]:
        pass

    @abstractmethod
    def preserve_create_edges(self, impulses: List[CreateEdgeImpulse]) -> Observable[CreatedEdgeImpulse]:
        pass

    @abstractmethod
    def preserve_delete_edges(self, impulses: List[DeleteEdgeImpulse]) -> Observable[DeletedEdgeImpulse]:
        pass

    @abstractmethod
    def perceive(self, config: Callable[[PerceptionReq], None]) -> Observable[PerceptionRes]:
        pass

    @abstractmethod
    def perceive_action(self, impulse: PerceiveActionImpulse) -> Observable[PerceivedImpulse]:
        pass

    @abstractmethod
    def perceive_data(self, impulse: PerceiveDataImpulse) -> Observable[PerceivedImpulse]:
        pass





from typing import Callable, List
from gaia_sdk.api.VariableRegistry import VariableRegistry
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType


class DeleteCodeImpulse():
    """
    The specification to delete a code instance
    """
    identityId: str
    reference: str

    def __init__(self, identityId: str, reference: str):
        self.identityId = identityId
        self.reference = reference

    def __eq__(self, other):
        if type(other) is type(self):
            return self.identityId == other.identityId and self.reference == other.reference
        return False

    def __repr__(self):
        return {'identityId': self.identityId, 'reference': self.reference}





from typing import Callable, List
from gaia_sdk.api.VariableRegistry import VariableRegistry
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType


class DeleteApiKeyImpulse():
    """
    The specification to delete an api key instance
    """
    apiKeyId: str

    def __init__(self, apiKeyId: str):
        self.apiKeyId = apiKeyId

    def __eq__(self, other):
        if type(other) is type(self):
            return self.apiKeyId == other.apiKeyId
        return False

    def __repr__(self):
        return {'apiKeyId': self.apiKeyId}





from typing import Callable, List
from gaia_sdk.api.VariableRegistry import VariableRegistry
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType


class CreateApiKeyImpulse():
    """
    The specification to create an api key instance
    """
    name: str
    description: str
    enabled: bool

    def __init__(self, name: str, description: str, enabled: bool):
        self.name = name
        self.description = description
        self.enabled = enabled

    def __eq__(self, other):
        if type(other) is type(self):
            return self.name == other.name and self.description == other.description and self.enabled == other.enabled
        return False

    def __repr__(self):
        return {'name': self.name, 'description': self.description, 'enabled': self.enabled}



import logging
import unittest

from rx import operators as ops, pipe

from gaia_sdk.api.DataRef import DataRefRequestConfig
from gaia_sdk.gaia import Gaia
from gaia_sdk.api.GaiaCredentials import UsernamePasswordCredentials
from gaia_sdk.http.response.FileListing import FileListing
from gaia_sdk.tests.mock import mock_gaia_ref
from gaia_sdk.tests.mock import MockResponse

logging.basicConfig(level=logging.DEBUG)


class TestDataRef(unittest.TestCase):

    def test_retrieve_data(self):
        def mock(request):
            self.assertEqual(request.url_post_fix, "/data/source")
            return MockResponse(bytes("hello world", encoding="utf-8"))

        self.gaiaRef = mock_gaia_ref(mock)
        result = pipe(ops.first())(
            self.gaiaRef.data("gaia://tenant/somefolder/somefolder/asdf1.pdf").as_bytes()).run()
        self.assertEqual(result, bytes("hello world", "utf-8"))

    def test_write_new_file(self):
        self.gaiaRef = mock_gaia_ref(self.mock_write)
        data = bytes("234", encoding="utf-8")
        response = self.gaiaRef.data("gaia://tenant/somefolder").add("newFile", data)
        assert pipe(ops.first())(response).run().uri == "gaia://tenant/somefolder/newFile"

    def test_overwrite_existing_file_doesnt_work(self):
        self.gaiaRef = mock_gaia_ref(self.mock_write_existing_fail)
        data = bytes("234", encoding="utf-8")
        self.assertRaises(Exception,
                          lambda: self.gaiaRef.data("gaia://tenant/somefolder").add("existingFile", data).pipe(
                              ops.first()).run())

    def test_write_new_file_progress(self):
        self.gaiaRef = mock_gaia_ref(self.mock_write)
        data = bytes("234", encoding="utf-8")

        def test_progress(value):
            assert value == 100

        config: DataRefRequestConfig = DataRefRequestConfig(test_progress)

        response = self.gaiaRef.data("gaia://tenant/somefolder").add("newFile", data, False, config)
        assert pipe(ops.first())(response).run().uri == "gaia://tenant/somefolder/newFile"

    def test_overwrite_file(self):
        self.gaiaRef = mock_gaia_ref(self.mock_write)
        data = bytes("234", encoding="utf-8")
        response = self.gaiaRef.data("gaia://tenant/somefolder").add("existingFile", data, override=True)
        assert pipe(ops.first())(response).run().uri == "gaia://tenant/somefolder/existingFile"

    def test_list_files_in_existing_dir(self):
        def mock(request):
            self.assertEqual(request.url_post_fix, "/data/list")
            return MockResponse([{"tenant": "tenant1", "filePath": "existingDirectory/file1",
                                  "lastModified": "2020-11-18", "size": "1000"}])

        self.gaiaRef = mock_gaia_ref(mock)
        result = pipe(ops.first())(self.gaiaRef.data("gaia://tenant1/existingDirectory").list()).run()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], FileListing({"tenant": "tenant1", "filePath": "existingDirectory/file1",
                                                 "lastModified": "2020-11-18", "size": "1000"}))

    @unittest.skip("E2E test, for local use or future E2E use")
    def test_list_files_local_e2e(self):
        self.gaiaRef = Gaia.login("http://localhost:8080", UsernamePasswordCredentials("admin", "admin"))

        result = pipe(ops.first())(
            self.gaiaRef.data("gaia://e3514269-a8a8-45df-8df8-851eccff6f55/cat1.jpg").list()).run()
        self.assertEqual(len(result), 14)

    def test_list_files_in_nonexistent_dir(self):
        def mock(request):
            self.assertEqual(request.url_post_fix, "/data/list")
            return MockResponse([])

        self.gaiaRef = mock_gaia_ref(mock)
        result = pipe(ops.first())(self.gaiaRef.data("gaia://tenant1/nonexistentDirectory").list()).run()
        self.assertEqual(len(result), 0)

    def test_remove_existing_file(self):
        def mock(request):
            self.assertEqual(request.url_post_fix, "/data/remove")
            return MockResponse({"fileExisted": True})

        self.gaiaRef = mock_gaia_ref(mock)
        result = pipe(ops.first())(self.gaiaRef.data("gaia://tenant/somefolder/existingFile").remove()).run()
        self.assertEqual(result.file_existed, True)

    def test_remove_nonexistent_file(self):
        def mock(request):
            self.assertEqual(request.url_post_fix, "/data/remove")
            return MockResponse({"fileExisted": False})

        self.gaiaRef = mock_gaia_ref(mock)
        result = pipe(ops.first())(self.gaiaRef.data("gaia://tenant/somefolder/nonexistentFile").remove()).run()
        self.assertEqual(result.file_existed, False)

    def mock_write(self, request):
        if request.url_post_fix == "/data/sink/init":
            return MockResponse({"uploadId": "A123"})
        elif request.url_post_fix == "/data/sink/chunk":
            return MockResponse({"chunkId": "B123"})
        elif request.url_post_fix == "/data/sink/complete":
            return MockResponse("asdfqwer")
        else:
            raise RuntimeError('On purpose')

    def mock_write_existing_fail(self, request):
        if request.url_post_fix == "/data/sink/init":
            raise RuntimeError('On purpose')
        else:
            return MockResponse({"x": "y"})



from rx.core.abc import Scheduler

from gaia_sdk.api.DataRef import DataRef
from gaia_sdk.api.ISensorStream import ISensorStream
from gaia_sdk.api.IdentityOp import IdentityOp
from gaia_sdk.api.SkillRef import SkillRef
from gaia_sdk.http.GaiaStreamClientBuilder import GaiaStreamClientFactory


class HttpSensorStream(ISensorStream):

    def __init__(self, url: str, credentials, scheduler: Scheduler, client_factory: GaiaStreamClientFactory):
        self._client = client_factory.http(url + "/api").with_credentials(credentials).build()
        self._scheduler = scheduler

    def data(self, uri: str):
        return DataRef(uri, self._client, self._scheduler)

    def identity(self):
        return IdentityOp(self._client, self._scheduler)

    def skill(self, uri: str) -> SkillRef:
        return SkillRef(uri, self._client, self._scheduler)




from gaia_sdk.graphql.request.type.BuildInEvaluation import BuildInEvaluation
from gaia_sdk.graphql.request.type.SkillEvaluation import SkillEvaluation

from typing import Callable, List
from gaia_sdk.api.VariableRegistry import VariableRegistry
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType


class Evaluation(list):

    def skill(self, config: Callable[['SkillEvaluation'], None]):
        def callback(registry: VariableRegistry):
            entity = SkillEvaluation()
            config(entity)
            return "skill {" + entity.render(registry) + "}"
        self.append(callback)

    def build_in(self, config: Callable[['BuildInEvaluation'], None]):
        def callback(registry: VariableRegistry):
            entity = BuildInEvaluation()
            config(entity)
            return "build_in {" + entity.render(registry) + "}"
        self.append(callback)

    def render(self, registry: VariableRegistry):
        return " ".join(map(lambda e: e(registry), self))




from gaia_sdk.graphql.request.type.Fulfilment import Fulfilment
from gaia_sdk.graphql.request.type.User import User
from gaia_sdk.graphql.request.type.ApiKey import ApiKey
from gaia_sdk.graphql.request.type.Behaviour import Behaviour
from gaia_sdk.graphql.request.type.Statement import Statement
from gaia_sdk.graphql.request.type.Intent import Intent
from gaia_sdk.graphql.request.type.Code import Code
from gaia_sdk.graphql.request.type.Role import Role
from gaia_sdk.graphql.request.type.SkillProvision import SkillProvision
from gaia_sdk.graphql.request.type.Skill import Skill
from gaia_sdk.graphql.request.type.Tenant import Tenant
from gaia_sdk.graphql.request.type.Prompt import Prompt
from gaia_sdk.graphql.request.type.Identity import Identity
from gaia_sdk.graphql.request.type.Edge import Edge

from typing import Callable, List
from gaia_sdk.api.VariableRegistry import VariableRegistry
from gaia_sdk.graphql.request.enumeration.Order import Order
from gaia_sdk.graphql.request.enumeration.OrderByField import OrderByField
from gaia_sdk.graphql.request.enumeration.EdgeOrderByField import EdgeOrderByField
from gaia_sdk.graphql.request.enumeration.EdgeType import EdgeType


class Knowledge(list):

    def users(self, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['User'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("limit", limit)
            name2 = registry.register("offset", offset)
            name3 = registry.register("orderBy", orderBy)
            name4 = registry.register("order", order)
            entity = User()
            config(entity)
            return f'users(limit:{name1}, offset:{name2}, orderBy:{name3}, order:{name4})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def user(self, userId: str, config: Callable[['User'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("userId", userId)
            entity = User()
            config(entity)
            return f'user(userId:{name1})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def api_keys(self, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['ApiKey'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("limit", limit)
            name2 = registry.register("offset", offset)
            name3 = registry.register("orderBy", orderBy)
            name4 = registry.register("order", order)
            entity = ApiKey()
            config(entity)
            return f'apiKeys(limit:{name1}, offset:{name2}, orderBy:{name3}, order:{name4})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def api_key(self, apiKeyId: str, config: Callable[['ApiKey'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("apiKeyId", apiKeyId)
            entity = ApiKey()
            config(entity)
            return f'apiKey(apiKeyId:{name1})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def roles(self, tenantId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Role'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("tenantId", tenantId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Role()
            config(entity)
            return f'roles(tenantId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def role(self, tenantId: str, roleId: str, config: Callable[['Role'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("tenantId", tenantId)
            name2 = registry.register("roleId", roleId)
            entity = Role()
            config(entity)
            return f'role(tenantId:{name1}, roleId:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def tenants(self, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Tenant'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("limit", limit)
            name2 = registry.register("offset", offset)
            name3 = registry.register("orderBy", orderBy)
            name4 = registry.register("order", order)
            entity = Tenant()
            config(entity)
            return f'tenants(limit:{name1}, offset:{name2}, orderBy:{name3}, order:{name4})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def tenant(self, tenantId: str, config: Callable[['Tenant'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("tenantId", tenantId)
            entity = Tenant()
            config(entity)
            return f'tenant(tenantId:{name1})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def identities(self, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Identity'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("limit", limit)
            name2 = registry.register("offset", offset)
            name3 = registry.register("orderBy", orderBy)
            name4 = registry.register("order", order)
            entity = Identity()
            config(entity)
            return f'identities(limit:{name1}, offset:{name2}, orderBy:{name3}, order:{name4})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def identity(self, identityId: str, config: Callable[['Identity'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            entity = Identity()
            config(entity)
            return f'identity(identityId:{name1})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def intents(self, identityId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Intent'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Intent()
            config(entity)
            return f'intents(identityId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def intent(self, identityId: str, reference: str, config: Callable[['Intent'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("reference", reference)
            entity = Intent()
            config(entity)
            return f'intent(identityId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def prompts(self, identityId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Prompt'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Prompt()
            config(entity)
            return f'prompts(identityId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def prompt(self, identityId: str, reference: str, config: Callable[['Prompt'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("reference", reference)
            entity = Prompt()
            config(entity)
            return f'prompt(identityId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def fulfilments(self, identityId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Fulfilment'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Fulfilment()
            config(entity)
            return f'fulfilments(identityId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def fulfilment(self, identityId: str, reference: str, config: Callable[['Fulfilment'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("reference", reference)
            entity = Fulfilment()
            config(entity)
            return f'fulfilment(identityId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def statements(self, identityId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Statement'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Statement()
            config(entity)
            return f'statements(identityId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def statement(self, identityId: str, reference: str, config: Callable[['Statement'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("reference", reference)
            entity = Statement()
            config(entity)
            return f'statement(identityId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def codes(self, identityId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Code'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Code()
            config(entity)
            return f'codes(identityId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def code(self, identityId: str, reference: str, config: Callable[['Code'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("reference", reference)
            entity = Code()
            config(entity)
            return f'code(identityId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def behaviours(self, identityId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Behaviour'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Behaviour()
            config(entity)
            return f'behaviours(identityId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def behaviour(self, identityId: str, reference: str, config: Callable[['Behaviour'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("identityId", identityId)
            name2 = registry.register("reference", reference)
            entity = Behaviour()
            config(entity)
            return f'behaviour(identityId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def edges(self, source: str, limit: int, offset: int, orderBy: EdgeOrderByField, order: Order, config: Callable[['Edge'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("source", source)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Edge()
            config(entity)
            return f'edges(source:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def edge(self, source: str, edgeId: str, config: Callable[['Edge'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("source", source)
            name2 = registry.register("edgeId", edgeId)
            entity = Edge()
            config(entity)
            return f'edge(source:{name1}, edgeId:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def skills(self, tenantId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['Skill'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("tenantId", tenantId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = Skill()
            config(entity)
            return f'skills(tenantId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def skill(self, tenantId: str, reference: str, config: Callable[['Skill'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("tenantId", tenantId)
            name2 = registry.register("reference", reference)
            entity = Skill()
            config(entity)
            return f'skill(tenantId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def skill_provisions(self, tenantId: str, limit: int, offset: int, orderBy: OrderByField, order: Order, config: Callable[['SkillProvision'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("tenantId", tenantId)
            name2 = registry.register("limit", limit)
            name3 = registry.register("offset", offset)
            name4 = registry.register("orderBy", orderBy)
            name5 = registry.register("order", order)
            entity = SkillProvision()
            config(entity)
            return f'skillProvisions(tenantId:{name1}, limit:{name2}, offset:{name3}, orderBy:{name4}, order:{name5})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def skill_provision(self, tenantId: str, reference: str, config: Callable[['SkillProvision'], None]):
        def callback(registry: VariableRegistry):
            name1 = registry.register("tenantId", tenantId)
            name2 = registry.register("reference", reference)
            entity = SkillProvision()
            config(entity)
            return f'skillProvision(tenantId:{name1}, reference:{name2})' + '{' + entity.render(registry) + '}'
        self.append(callback)

    def render(self, registry: VariableRegistry):
        return " ".join(map(lambda e: e(registry), self))



import logging

import requests
from requests import Response

from gaia_sdk.api.client_options import ClientOptions
from gaia_sdk.api.transporter.abstract_transporter import ITransporter
from gaia_sdk.http.request.Payload import Payload


class HttpTransporter(ITransporter):
    """
    ITransporter implementation which is based on the HTTP protocol.
    """

    def __init__(self, url):
        self.url = url
        self.logger = logging.getLogger("HttpTransporter")

    def transport(self, options: ClientOptions, payload: Payload, url_post_fix: str = "") -> Response:
        if payload.payload_type == Payload.STREAM:
            data = payload.binaryData
        else:
            data = payload.data

        headers = HttpTransporter.get_default_headers(options, data)
        url = self.url + url_post_fix
        self.logger.debug("request to %s header:%s payload:%r", url, headers, data)
        if payload.payload_type == Payload.JSON:
            response = requests.post(url, json=data, headers=headers)
        elif payload.payload_type == Payload.FORM_DATA:
            response = requests.post(url, files=data, headers=headers)
        elif payload.payload_type == Payload.STREAM:
            response = requests.post(url, data=data, headers=headers, params=options.request_parameters)
        else:
            raise ValueError(f"Unsupported type for payload {payload.payload_type}")
        self.logger.debug("response: %r", response)
        if response.status_code != 200:
            raise self.__raise_for_status_code(unsuccessful_response=response)
        return response

    def __raise_for_status_code(self, unsuccessful_response: Response) -> Exception:

        def with_message(response: Response, message: str):
            return Exception(f"[{response.status_code}] - " + message)

        if unsuccessful_response.status_code == 401:
            return with_message(unsuccessful_response,
                                f"Could not authenticate with '{unsuccessful_response.url}': {unsuccessful_response.text}")
        if unsuccessful_response.status_code == 500:
            return with_message(unsuccessful_response, f"Server error: '{unsuccessful_response.text}'")
        return with_message(unsuccessful_response, f"Unknown error occurred: {unsuccessful_response.text}")

    @staticmethod
    def build_authorization_header(options: ClientOptions, payload: dict) -> str:
        if options.credentials is None:
            raise ValueError("Authorization Header cannot be generated because no credentials are set")
        return options.credentials.create_auth_header(options, payload)

    @staticmethod
    def get_default_headers(options: ClientOptions, payload: dict) -> dict:
        if options.credentials is None:
            raise ValueError("Authorization Header cannot be generated because no credentials are set")
        return {"Authorization": options.credentials.create_auth_header(options, payload)}
